#!/usr/bin/env node

'use strict';

var commander = require('commander');
var path = require('path');
var rxjs = require('rxjs');
var operators = require('rxjs/operators');
var fs = require('fs');
var inquirer = require('inquirer');

function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () {
                        return e[k];
                    }
                });
            }
        });
    }
    n['default'] = e;
    return Object.freeze(n);
}

var path__namespace = /*#__PURE__*/_interopNamespace(path);
var fs__namespace = /*#__PURE__*/_interopNamespace(fs);
var inquirer__namespace = /*#__PURE__*/_interopNamespace(inquirer);

const isBrowser = () => typeof window !== 'undefined';

// NOTE: use bind(window) to avoid `TypeError: Illegal invocation`
const fetch = isBrowser() ? globalThis.fetch.bind(window) : require('node-fetch');
class DefaultHttpClient {
    fetchFn;
    constructor(fetchFn = fetch) {
        this.fetchFn = fetchFn;
    }
    async get(url, options = {}) {
        const result = await this.fetchFn(url, options);
        return (await result.json());
    }
}

const baseURL = 'https://scrapbox.io/api';
const isObject = (data) => typeof data === 'object' && data !== null;
const isNotLoggedInError = (data) => isObject(data) && data['name'] === 'NotLoggedInError';
class RestApiClient {
    token;
    httpClient;
    constructor(token, httpClient = new DefaultHttpClient()) {
        this.token = token;
        this.httpClient = httpClient;
    }
    async getPage(projectName, pageName) {
        return this.request(`${baseURL}/pages/${encodeURIComponent(projectName)}/${encodeURIComponent(pageName)}?followRename=true`);
    }
    async getProject(projectName) {
        return this.request(`${baseURL}/projects/${encodeURIComponent(projectName)}`);
    }
    async getProjects() {
        return this.request(`${baseURL}/projects`);
    }
    async getMe() {
        return this.request(`${baseURL}/users/me`);
    }
    async request(url) {
        const result = await this.httpClient.get(url, {
            headers: {
                Cookie: `connect.sid=${this.token}`,
            },
        });
        if (isNotLoggedInError(result)) {
            throw new Error('NotLoggedInError');
        }
        return result;
    }
}

const packetTypes = {
    initialize: '0',
    ping: '2',
    pong: '3',
    connected: '40',
    // e.g.
    //   * 42: 'cursor'
    //   * 42X: user custom request (X is arbitrary natural number to specify response)
    send: '42',
    // e.g.
    //   * 43X: response to user custom request (X is number used for request)
    receive: '43',
    response: '43',
};
const CONFIG = {
    endpoint: 'wss://scrapbox.io/socket.io/?EIO=4&transport=websocket',
    origin: 'https://scrapbox.io',
    responseTimeout: 1000 * 15,
};

const isIntegerString = (value) => Number.isInteger(Number.parseInt(value));
const getCookieToAuth = (token) => `connect.sid=${token}`;

const toSocketIoMessagePayload = (id, data) => {
    return `${packetTypes.send}${id}${JSON.stringify(['socket.io-request', data])}`;
};
// 430[{...}}] => 430, [{}]
// @see https://github.com/socketio/engine.io-protocol
const parseMessage = (message) => {
    let packetType = '';
    while (message.length) {
        const head = message[0];
        if (isIntegerString(head)) {
            packetType += head;
            message = message.substr(1);
        }
        else {
            return [packetType, JSON.parse(message)];
        }
    }
    return [packetType, undefined];
};
const isConnectionMessage = (message) => {
    return message[0] === packetTypes.initialize;
};
const isResponseMessage = (message) => {
    return message[0].length > 2 && message[0].startsWith(packetTypes.response);
};
const getRequestId = (packetType) => packetType.slice(2);
const isResponseMessageOf = (sid) => (message) => {
    return isResponseMessage(message) && sid === getRequestId(message[0]);
};

// copied from official index.js
// modified original toString16 of official index.js
const toString16 = (userIdOrNumber, length) => {
    const hexOrId = userIdOrNumber.toString(16);
    return hexOrId.length > length ? hexOrId.slice(hexOrId.length - length) : '0'.repeat(length - hexOrId.length) + hexOrId;
};
const map = new Map();
const getIdGenerator = (userId) => {
    if (map.has(userId)) {
        return map.get(userId);
    }
    let counter = Math.floor(16777215 * Math.random());
    const generate = () => {
        const t = Math.floor(Date.now() / 1e3);
        if ((counter += 1) > 16777215) {
            counter = 0;
        }
        return (toString16(t, 8) + toString16(userId, 6) + toString16(0, 4) + toString16(counter, 6));
    };
    map.set(userId, generate);
    return generate;
};
const generateId = (userId) => {
    const generator = getIdGenerator(userId);
    return generator();
};

// NOTE: userId is used to generate new ID
const createChanges = (params, userId) => params.map((param) => {
    switch (param.type) {
        case 'insert': {
            return { _insert: param.position || '_end', lines: { id: generateId(userId), text: param.text } };
        }
        case 'update': {
            return { _update: param.id, lines: { text: param.text } };
        }
        case 'delete': {
            return { _delete: param.id, lines: -1 };
        }
        case 'title': {
            return { title: param.title };
        }
        case 'description': {
            return { descriptions: [param.text] };
        }
        default: {
            throw new Error('Bad implement');
        }
    }
});

const getBrowserWebSocket = ({ endpoint }) => new WebSocket(endpoint);
const getNodeJSWebSocket = ({ endpoint, origin, cookie }) => {
    const NodeWebsocket = require('ws');
    const headers = { Origin: origin, Cookie: cookie };
    return new NodeWebsocket(endpoint, undefined, { headers });
};
const getWebSocket = (settings) => {
    if (isBrowser()) {
        // @TODO: validate settings
        return getBrowserWebSocket(settings);
    }
    else {
        // @TODO: validate settings
        return getNodeJSWebSocket(settings);
    }
};
const getEventStream = (socket) => {
    const close = rxjs.fromEvent(socket, 'close').pipe(operators.first());
    const open = rxjs.fromEvent(socket, 'open').pipe(operators.takeUntil(close), operators.shareReplay());
    const error = rxjs.fromEvent(socket, 'error').pipe(operators.takeUntil(close), operators.shareReplay());
    const message = rxjs.fromEvent(socket, 'message').pipe(operators.takeUntil(close));
    return { open, close, error, message };
};

class WebsocketClient {
    config;
    socket;
    open$;
    error$;
    close$;
    message$;
    // currently joined-room, need cache to re-join on reconnect
    room = null;
    // wait request until connection opened
    sid = 0;
    constructor(config = {}) {
        this.config = config;
        this.initialize();
    }
    commit(params) {
        return this.send({
            method: 'commit',
            data: {
                kind: 'page',
                userId: params.userId,
                projectId: params.projectId,
                pageId: params.pageId,
                parentId: params.parentId,
                changes: createChanges(params.changes, params.userId),
                cursor: null,
                freeze: true,
            },
        });
    }
    join(params) {
        this.room = params;
        return this.send({
            method: 'room:join',
            data: {
                pageId: params.pageId,
                projectId: params.projectId,
                projectUpdatesStream: false,
            },
        });
    }
    initialize() {
        this.socket = this.openWebSocket();
        const { open, error, close, message } = getEventStream(this.socket);
        this.open$ = open.pipe(operators.shareReplay());
        this.error$ = error.pipe(operators.shareReplay());
        this.close$ = close.pipe(operators.shareReplay());
        this.message$ = message.pipe(operators.map((event) => parseMessage(event.data)));
        this.registerInitializationOnOpen();
        this.registerReconnectionOnClose();
    }
    async send(data) {
        const sid = String(this.sid++);
        const message = toSocketIoMessagePayload(sid, data);
        this.socket.send(message);
        return rxjs.firstValueFrom(this.message$.pipe(operators.first(isResponseMessageOf(sid)), operators.timeout(CONFIG.responseTimeout)));
    }
    openWebSocket() {
        const getSocket = this.config.websocketConstructor ?? getWebSocket;
        return getSocket({ ...CONFIG, cookie: getCookieToAuth(this.config.token ?? '') });
    }
    registerInitializationOnOpen() {
        this.message$.pipe(operators.first(isConnectionMessage)).subscribe(([, data]) => {
            this.socket.send(packetTypes.connected);
            // start ping
            rxjs.interval(data.pingInterval)
                .pipe(operators.takeUntil(this.close$))
                .subscribe(() => this.socket.send(packetTypes.ping));
            // for reconnection after closed
            if (this.room) {
                this.join(this.room);
            }
        });
    }
    registerReconnectionOnClose() {
        this.close$.subscribe(() => this.initialize());
    }
}

class ScrapboxClient {
    apiClient;
    websocketClient;
    constructor(token = '', apiClient = new RestApiClient(token), websocketClient = new WebsocketClient({ token })) {
        this.apiClient = apiClient;
        this.websocketClient = websocketClient;
    }
    async getPage(projectName, pageName) {
        return this.apiClient.getPage(projectName, pageName);
    }
    async changeLine(projectName, pageName, change) {
        const project = await this.getProject(projectName);
        const { pageId, commitId } = await this.getPageIdAndCommitId(projectName, pageName);
        return this.commit({
            changes: Array.isArray(change) ? change : [change],
            projectId: project.id,
            pageId: pageId,
            commitId: commitId,
        });
    }
    async getUser() {
        return await this.apiClient.getMe();
    }
    async getProject(projectName) {
        return await this.apiClient.getProject(projectName);
    }
    async getPageIdAndCommitId(projectName, pageName) {
        const page = await this.apiClient.getPage(projectName, pageName);
        return { pageId: page.id, commitId: page.commitId };
    }
    async commit(param) {
        const user = await this.getUser();
        // to receive result of commit
        await this.websocketClient.join({ projectId: param.projectId, pageId: param.pageId });
        return this.websocketClient.commit({
            userId: user.id,
            projectId: param.projectId,
            pageId: param.pageId,
            parentId: param.commitId,
            changes: param.changes,
        });
    }
}

const loadFile = (path) => new Promise((resolve, reject) => fs__namespace.readFile(path, 'utf8', (err, data) => {
    if (err) {
        reject(err);
    }
    resolve(data);
}));
const getJSTString = () => new Date().toLocaleString('ja-JP', { timeZone: 'Asia/Tokyo' });
const versionComment = `/* deployed at: ${getJSTString()} */`;
// first space is needed for scrapbox code block
const loadSourceCode = async (file) => ` ${versionComment}${await loadFile(file)}`;

const findIndex = (searchString, lines) => {
    return lines.findIndex((line) => line.text.includes(searchString));
};
const findNextLineIdOrFail = (searchString, lines) => {
    const index = findIndex(searchString, lines);
    if (index === -1) {
        throw new Error('Line not found');
    }
    const nextLineOrNone = lines[index + 1];
    if (!nextLineOrNone) {
        throw new Error('Next line not found');
    }
    return nextLineOrNone.id;
};

const isNonEmptyString = (value) => typeof value === 'string' && value.length > 0;
const isCSSFile = (value) => value.endsWith('.css');
const isJSFile = (value) => value.endsWith('.js');
const validateDeployArguments = ({ project, page, sourceFilePath, token }) => {
    const errors = [];
    if (!isNonEmptyString(project)) {
        errors.push('project is required');
    }
    if (!isNonEmptyString(page)) {
        errors.push('page is required');
    }
    if (!isNonEmptyString(sourceFilePath)) {
        errors.push('filepath is required');
    }
    if (!(isCSSFile(sourceFilePath) || isJSFile(sourceFilePath))) {
        errors.push('file type is not css or js');
    }
    if (!isNonEmptyString(token)) {
        errors.push('token is required');
    }
    return errors;
};

const deploy = async ({ page, sourceFilePath, project, token }) => {
    const errors = validateDeployArguments({ token, sourceFilePath, project, page });
    if (errors.length) {
        throw new Error(`ParameterError: ${JSON.stringify(errors)}`);
    }
    const codeBlockLabel = isCSSFile(sourceFilePath) ? 'style.css' : 'script.js';
    const sourceCode = await loadSourceCode(sourceFilePath);
    const scrapboxClient = new ScrapboxClient(token);
    const pageData = await scrapboxClient.getPage(project, page);
    // NOTE: assuming code-block exists already
    const lineId = findNextLineIdOrFail(codeBlockLabel, pageData.lines);
    // FIXME: bad type assertion
    await scrapboxClient.changeLine(project, page, { type: 'update', id: lineId, text: sourceCode });
};

const selectProject = async ({ client }) => {
    const projectOptions = await client.getProjects().then(({ projects }) => projects.map(({ name }) => name));
    const answers = await inquirer__namespace.prompt([{ type: 'list', name: 'project', message: 'select target project', choices: projectOptions }]);
    return answers.project;
};

const pkg = require(path__namespace.resolve(__dirname, '../../package.json'));
const program = new commander.Command().version(pkg.version);
program
    .command('deploy-userscript <js-file>')
    .description('update userscript in user page')
    .option('-p, --project <type>', 'project name')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApiClient(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    const userName = await api.getMe().then(({ name }) => name);
    await deploy({ token, project: selectedProject, page: userName, sourceFilePath: jsFile });
});
program
    .command('deploy-usercss <css-file>')
    .description('update usercss in settings page')
    .option('-p, --project <type>', 'project name')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApiClient(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    await deploy({ token, project: selectedProject, page: 'settings', sourceFilePath: jsFile });
});
program
    .parseAsync(process.argv)
    .then(() => {
    process.exit(0);
})
    .catch((e) => {
    console.error(e);
    process.exit(1);
});
//# sourceMappingURL=bin.js.map
