'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var rxjs = require('rxjs');
var operators = require('rxjs/operators');
var constants = require('./constants.js');
var message = require('./internal/message.js');
var request = require('./internal/request.js');
var util = require('./internal/util.js');
var websocket = require('./internal/websocket.js');

class WebsocketClient {
  config;
  socket;
  open$;
  error$;
  close$;
  message$; // currently joined-room, need cache to re-join on reconnect

  room = null; // wait request until connection opened

  sid = 0;

  constructor(config = {}) {
    this.config = config;
    this.initialize();
  }

  commit(params) {
    return this.send({
      method: 'commit',
      data: {
        kind: 'page',
        userId: params.userId,
        projectId: params.projectId,
        pageId: params.pageId,
        parentId: params.parentId,
        changes: request.createChanges(params.changes, params.userId),
        cursor: null,
        freeze: true
      }
    });
  }

  join(params) {
    this.room = params;
    return this.send({
      method: 'room:join',
      data: {
        pageId: params.pageId,
        projectId: params.projectId,
        projectUpdatesStream: false
      }
    });
  }

  initialize() {
    this.socket = this.openWebSocket();
    const {
      open,
      error,
      close,
      message: message$1
    } = websocket.getEventStream(this.socket);
    this.open$ = open.pipe(operators.shareReplay());
    this.error$ = error.pipe(operators.shareReplay());
    this.close$ = close.pipe(operators.shareReplay());
    this.message$ = message$1.pipe(operators.map(event => message.parseMessage(event.data)));
    this.registerInitializationOnOpen();
    this.registerReconnectionOnClose();
  }

  async send(data) {
    const sid = String(this.sid++);
    const message$1 = message.toSocketIoMessagePayload(sid, data);
    this.socket.send(message$1);
    return rxjs.firstValueFrom(this.message$.pipe(operators.first(message.isResponseMessageOf(sid)), operators.timeout(constants.CONFIG.responseTimeout)));
  }

  openWebSocket() {
    const getSocket = this.config.websocketConstructor ?? websocket.getWebSocket;
    return getSocket({ ...constants.CONFIG,
      cookie: util.getCookieToAuth(this.config.token ?? '')
    });
  }

  registerInitializationOnOpen() {
    this.message$.pipe(operators.first(message.isConnectionMessage)).subscribe(([, data]) => {
      this.socket.send(constants.packetTypes.connected); // start ping

      rxjs.interval(data.pingInterval).pipe(operators.takeUntil(this.close$)).subscribe(() => this.socket.send(constants.packetTypes.ping)); // for reconnection after closed

      if (this.room) {
        this.join(this.room);
      }
    });
  }

  registerReconnectionOnClose() {
    this.close$.subscribe(() => this.initialize());
  }

}

exports.WebsocketClient = WebsocketClient;
//# sourceMappingURL=websocket-client.js.map
