import { firstValueFrom, interval } from 'rxjs';
import { shareReplay, map, first, timeout, takeUntil } from 'rxjs/operators';
import { CONFIG, packetTypes } from './constants.js';
import { parseMessage, toSocketIoMessagePayload, isResponseMessageOf, isConnectionMessage } from './internal/message.js';
import { createChanges } from './internal/request.js';
import { getCookieToAuth } from './internal/util.js';
import { getEventStream, getWebSocket } from './internal/websocket.js';

class WebsocketClient {
    config;
    socket;
    open$;
    error$;
    close$;
    message$;
    // currently joined-room, need cache to re-join on reconnect
    room = null;
    // wait request until connection opened
    sid = 0;
    constructor(config = {}) {
        this.config = config;
        this.initialize();
    }
    commit(params) {
        return this.send({
            method: 'commit',
            data: {
                kind: 'page',
                userId: params.userId,
                projectId: params.projectId,
                pageId: params.pageId,
                parentId: params.parentId,
                changes: createChanges(params.changes, params.userId),
                cursor: null,
                freeze: true,
            },
        });
    }
    join(params) {
        this.room = params;
        return this.send({
            method: 'room:join',
            data: {
                pageId: params.pageId,
                projectId: params.projectId,
                projectUpdatesStream: false,
            },
        });
    }
    initialize() {
        this.socket = this.openWebSocket();
        const { open, error, close, message } = getEventStream(this.socket);
        this.open$ = open.pipe(shareReplay());
        this.error$ = error.pipe(shareReplay());
        this.close$ = close.pipe(shareReplay());
        this.message$ = message.pipe(map((event) => parseMessage(event.data)));
        this.registerInitializationOnOpen();
        this.registerReconnectionOnClose();
    }
    async send(data) {
        const sid = String(this.sid++);
        const message = toSocketIoMessagePayload(sid, data);
        this.socket.send(message);
        return firstValueFrom(this.message$.pipe(first(isResponseMessageOf(sid)), timeout(CONFIG.responseTimeout)));
    }
    openWebSocket() {
        const getSocket = this.config.websocketConstructor ?? getWebSocket;
        return getSocket({ ...CONFIG, cookie: getCookieToAuth(this.config.token ?? '') });
    }
    registerInitializationOnOpen() {
        this.message$.pipe(first(isConnectionMessage)).subscribe(([, data]) => {
            this.socket.send(packetTypes.connected);
            // start ping
            interval(data.pingInterval)
                .pipe(takeUntil(this.close$))
                .subscribe(() => this.socket.send(packetTypes.ping));
            // for reconnection after closed
            if (this.room) {
                this.join(this.room);
            }
        });
    }
    registerReconnectionOnClose() {
        this.close$.subscribe(() => this.initialize());
    }
}

export { WebsocketClient };
//# sourceMappingURL=websocket-client.js.map
